from urllib.request import urlopen, Request, urljoin
from base64 import b64decode

import json

from .exceptions import AdonisAuthenticationDeniedError, AdonisError
from .crew import Crew
from ..const import USER_AGENT

from PIL.Image import Image


CREW_PORTAL = 0
INTEGRATION = 1


class Adonis:
    def __init__(
        self,
        crew_portal_base_url: str,
        integration_base_url: str,
        login: str,
        password: str,
    ):
        self.crew_portal_base_url = crew_portal_base_url
        self.integration_base_url = integration_base_url
        self.login = login
        self.password = password

    @classmethod
    def fromConfig(cls, config):
        return cls(
            config.crew_portal_base_url,
            config.integration_base_url,
            config.login,
            config.password,
        )

    def request(
        self,
        endpoint: str,
        payload: dict,
        add_token: bool = True,
        wrap_request: bool = True,
        extract_response: bool = True,
        api: int = CREW_PORTAL,
    ):
        base_url = (
            self.crew_portal_base_url
            if api == CREW_PORTAL
            else self.integration_base_url
        )
        url = urljoin(base_url, endpoint)

        req = Request(url)
        req.add_header("User-Agent", USER_AGENT)
        req.add_header("Content-Type", "application/json")
        req.add_header("Accept", "application/json")

        if add_token:
            payload["Authentication_Token"] = self.getToken(api=api)

        if wrap_request:
            payload = {"request": payload}

        req.data = json.dumps(payload).encode()

        res = json.loads(urlopen(req).read())

        if extract_response:
            res = res[endpoint + "Result"]

        if not res["Authentication_Approved"]:
            raise AdonisAuthenticationDeniedError(res["Authentication_ReasonDenied"])

        if res["ErrorText"]:
            raise AdonisError(res["ErrorText"])

        return res

    def requestIntegration(
        self,
        endpoint: str,
        payload: dict,
        add_token: bool = True,
        wrap_request: bool = True,
        extract_response: bool = True,
    ):
        return self.request(
            endpoint, payload, add_token, wrap_request, extract_response, INTEGRATION
        )

    def requestCrewPortal(
        self,
        endpoint: str,
        payload: dict,
        add_token: bool = True,
        wrap_request: bool = True,
        extract_response: bool = True,
    ):
        return self.request(
            endpoint, payload, add_token, wrap_request, extract_response, CREW_PORTAL
        )

    def getToken(self, lifetime: int = 60, api: int = CREW_PORTAL):
        data = {
            "credentials": {
                "Login": self.login,
                "Password": self.password,
                "LifeTime": lifetime,
            }
        }

        method = (
            self.requestCrewPortal if api == CREW_PORTAL else self.requestIntegration
        )
        result = method("GNL_API_AUTHENTICATION", data, False, False)

        return result["Authentication_Token"]

    def getCrewPersonalData(self, pin):
        data = {"Pin": pin}

        result = self.requestCrewPortal("DG_GeneralDetailsRead", data)

        return result

    def getCrewProfilePicture(self, pin):
        data = {"Pin": pin}

        try:
            result = self.requestCrewPortal("DG_ProfilePictureRead", data)
        except AdonisAuthenticationDeniedError:
            raise
        except AdonisError:
            return None

        return result["FileBase64"]

    def getCrew(self, pin, personal_data=True, profile_picture=True):
        crew = Crew(pin)

        if personal_data:
            crew._personal_data = self.getCrewPersonalData(pin)

        if profile_picture:
            crew._profile_picture = self.getCrewProfilePicture(pin)

        return crew

    def getCrewListView(self, view, filters="", pagination="", paginate_by="", page=""):
        data = {
            "View": view,
            "Filter": filters,
            "Pagination": pagination or "None",
            "RowsByPage": paginate_by,
            "Page": page,
        }

        try:
            result = self.requestIntegration("GNL_APMCrewListViews", data)
        except AdonisAuthenticationDeniedError:
            raise
        except AdonisError:
            result = None

        return result

    def getRequirements(self, view, paginate_by=100):
        requirements = []
        pnum = 1

        while True:
            page = self.getCrewListView(
                view, pagination="True", paginate_by=str(paginate_by), page=str(pnum)
            )

            for record in json.loads(page["Result"]):
                requirements.append((record["PIN"], record["Document code"]))

            if page["CurrentPage"] == page["TotalPages"]:
                break

            pnum += 1

        return requirements

    def getCodes(self):
        try:
            return self.requestIntegration("DG_GetCodes")
        except AdonisError:
            return None

    def writeCompetence(
        self,
        pin,
        competence_code,
        certificate_number,
        from_timestamp=None,
        to_timestamp=None,
        issued_by="pyadonis",
        attachment=None,
        **kwargs,
    ):
        from_dt = (
            datetime.datetime.fromtimestamp(from_timestamp)
            if from_timestamp
            else datetime.datetime.now()
        )
        from_dt = from_dt.replace(hour=0, minute=0, second=0, microsecond=0)
        from_dt_string = from_dt.isoformat()

        if to_timestamp:
            to_dt = datetime.datetime.fromtimestamp(to_timestamp)
        else:
            codes = self.getCodes()
            expiry_days = 0

            for competence in codes["Competence"]:
                if competence["Code"] == competence_code:
                    expiry_days = int(competence["DefaultExpiryDays"])

            if expiry_days == 0:
                to_dt_string = ""
            else:
                if not expiry_days % 365:
                    to_dt = from_dt + relativedelta(years=expiry_days / 365)
                elif not expiry_days % 30:
                    to_dt = from_dt + relativedelta(months=expiry_days / 30)
                else:
                    to_dt = from_dt + relativedelta(days=expiry_days)

                to_dt_string = to_dt.isoformat()

        payload = {
            "Certificate": {
                "Pin": pin,
                "Code": competence_code,
                "Number": certificate_number,
                "DateFrom": from_dt_string,
                "ExpiryDate": to_dt_string,
                "IssuedBy": issued_by,
                "Planned": "False",
            }
        }

        if attachment:
            payload["Certificate"]["FileName"] = "certificate.pdf"

            if isinstance(attachment, bytes):
                attachment = attachment.decode()

            payload["Certificate"]["DocumentBase64"] = (
                "data:application/pdf;" + attachment
            )

        payload.update(kwargs)

        return self.requestCrewPortal("DG_CertificateCreate", payload)

    def getCompetences(self, pin):
        payload = {"Pin": pin}

        return self.requestCrewPortal("DG_CertificateRead", payload)
